/*
 * This is the High Performance Python Profiler portion of HotShot.
 *
 * This is an enhanced version of source file .../Modules/_hotshot.c
 * from Python release 2.5.2 using the high-resolution timer where
 * available and otherwise falling back to gettimeofday as before.
 *
 * All profile times are recorded in 64-bit ticks.  The tick length
 * can be obtained from new info item 'timer-frequency' logged as
 * ticks per second.  Two other files .../Lib/hotshot/log.py and
 * .../Lib/hotshot/stats.py have been updated and use the new info
 * item 'timer-frequency'.
 *
 * The high-resolution timer is used only if the C compiler does
 * support 64-bit long, either as PY_LONG_LONG or if SIZEOF_LONG
 * equals 8.  Several changes were made to log (delta) ticks as 64-
 * bit longs when needed.  In addition, this version avoids using
 * malloc where possible.
 *
 * Tested with Python 2.5.2 on 32-bit MacOS X (both Intel and PPC),
 * 32-bit CentOS, 64-bit RHEL 3u7 and 64-bit Solaris 10.
 *
 * A high-resolution timer may be useful elsewhere.  Perhaps as
 * another function in the time module and/or in a core file like
 * .../Python/ceval.c.
 */

#include "Python.h"
#include "code.h"
#include "eval.h"
#include "frameobject.h"
#include "structmember.h"

/*
 * Which timer to use should be made more configurable,
 * that should not be difficult.  This will do for now.
 */

#ifndef inline
# if defined(__GNUC__) && !defined(__cplusplus)
#   define inline __inline
# else
#   define inline
# endif
#endif

#ifdef MS_WINDOWS
# include <windows.h>

# ifdef HAVE_DIRECT_H
#   include <direct.h>    /* for getcwd() */
# endif

    /* XXX assume HAVE_LONG_LONG */
    typedef __int64 _ticks_t;

    static inline _ticks_t
    _gethrticks (void)
    {
        LARGE_INTEGER t; \
        QueryPerformanceCounter(&t); \
        return ((_ticks_t) t.QuadPart);
    }

#   define _TICKSIZE   0  /* no gettimeofday */
#   define _RUSAGE  0  /* no getrusage */

#else
  /* use hires timer only if 64-bit long type
     is available, otherwise fall back to long
     (and use _gettimeofday as timer) */
# ifdef HAVE_LONG_LONG
#   define _TICKSIZE  2
#   define _ticks_t   unsigned PY_LONG_LONG
# elif SIZEOF_LONG = 8
#   define _TICKSIZE  1
#   define _ticks_t   unsigned long
#else
#   define _TICKSIZE  0
#   define _ticks_t   long
# endif

# if defined(PYOS_OS2) || defined(__QNX__) || defined(__VMS)
#   include <sys/time.h>
#   define _RUSAGE  0

# else
#   include <sys/resource.h>
#   include <sys/times.h>
#   define _RUSAGE  1

    static _ticks_t
    _getrusage (void)
    {
        struct rusage ru;
        getrusage(RUSAGE_SELF, &ru);
        return ((_ticks_t) (ru.ru_utime.tv_usec + ru.ru_stime.tv_usec) + (_ticks_t) 1000000
              * (_ticks_t) (ru.ru_utime.tv_sec  + ru.ru_stime.tv_sec));
    }
# endif

# ifndef HAVE_GETTIMEOFDAY
#   error "Need gettimeofday() function in _hotshot.c"
# endif
    static inline _ticks_t
    _gettimeofday (void)  /* in usec ticks */
    {
        struct timeval tv;
#     ifdef GETTIMEOFDAY_NO_TZ
        gettimeofday(&tv)
#     else
        gettimeofday(&tv, (struct timezone *) NULL);
#     endif
        return ((_ticks_t) tv.tv_usec + (_ticks_t) 1000000 * (_ticks_t) tv.tv_sec);
    }

  /* XXX HAVE_GETHRTIME should probably be defined
     in pyconfig.h.  Both Solaris and HP-UX provide
     function gethrtime returning a time stamp with
     nanosecond accuracy (and using a very fast
     system call on UltraSPARC) */
# if _TICKSIZE && defined(HAVE_GETHRTIME)  /* Solaris and HP-UX only */
    /* XXX tested */
#   include <time.h>
#   define  _gethrticks  (_ticks_t) gethrtime

# elif _TICKSIZE && defined(__GNUC__) && (defined(__i386__) || defined(__x86_64__))
    static inline _ticks_t
    _gethrticks (void)
    {
       _ticks_t t;
#     ifdef rdtscll
         /* rdtscll() is defined in <asm/msr.h>
            of certain, recent Linux distro's */
        rdtscll(t);
#     else
        /* XXX tested */
        unsigned int hi, lo;
        asm volatile ("rdtsc" : "=a" (lo), "=d" (hi));
        t = (((_ticks_t) hi) << 32) | (_ticks_t) lo;
#     endif
        return (t);
    }

# elif _TICKSIZE && defined (__GNUC__) && (defined(__powerpc64__) || defined(__POWERPC64__))
    /* XXX untested */
    static /* inline */ _ticks_t
    _gethrticks (void)
    {
       _ticks_t t;
        asm volatile ("mftb %0" : "=r" (t));
        return (t);
    }

# elif _TICKSIZE && defined (__GNUC__) && (defined(__ppc__) || defined(__POWERPC__))
    /* XXX tested */
    static /* inline */ _ticks_t
    _gethrticks (void)
    {
       _ticks_t t;
        unsigned int hi, hi2, lo;
        do {  /* see also .../Python/ceval.c */
            asm volatile ("mftbu %0 \n"
                          "mftb  %1 \n"
                          "mftbu %2 \n" : "=r"(hi), "=r"(lo), "=r"(hi2));
	    } while (__builtin_expect(hi != hi2, 0));
#     if 0  /* XXX tested, see also .../Python/ceval.c */
	    ((unsigned int*)(t))[0] = hi;
	    ((unsigned int*)(t))[1] = lo;
#     else  /* XXX tested */
        t = (((_ticks_t) hi) << 32) | (_ticks_t) lo;
#     endif
        return (t);
    }

# else  /* no hires timer or 64-bit long */
#   undef   _TICKSIZE
#   define  _TICKSIZE  0
#   undef   _ticks_t
#   define  _ticks_t   long
    /* use _gettimeofday returning usecs */
#   define _gethrticks  _gettimeofday
# endif
#endif

#define BUFFERSIZE 10240

#if defined(PYOS_OS2) && defined(PYCC_GCC)
#define PATH_MAX 260
#endif

#if defined(__sgi) && _COMPILER_VERSION>700 && !defined(PATH_MAX)
/* fix PATH_MAX not being defined with MIPSPro 7.x
   if mode is ANSI C (default) */
#define PATH_MAX 1024
#endif

#ifndef PATH_MAX
#   ifdef MAX_PATH
#       define PATH_MAX MAX_PATH
#   elif defined (_POSIX_PATH_MAX)
#       define PATH_MAX _POSIX_PATH_MAX
#   else
#       error "Need a defn. for PATH_MAX in _hotshot.c"
#   endif
#endif

typedef struct {
    PyObject_HEAD
    PyObject *filemap;
    PyObject *logfilename;
    Py_ssize_t index;
    unsigned char buffer[BUFFERSIZE];
    FILE *logfp;
    int lineevents;
    int linetimings;
    int frametimings;
    /* size_t filled; */
    int active;
    int next_fileno;
   _ticks_t prev_ticks;
} ProfilerObject;

typedef struct {
    PyObject_HEAD
    PyObject *info;
    FILE *logfp;
    int linetimings;
    int frametimings;
} LogReaderObject;


static PyObject *ProfilerError = NULL;

static double timer_frequency  = 0.0;
static long   timer_resolution = 0L;
static long   timeofday_diff   = 0L;
static long   rusage_diff      = 0L;  /* -1L if _getrusage n/a */

#if _TICKSIZE
  /* start ticks and clock time */
  static _ticks_t ticks0       = (_ticks_t) 0;
  static _ticks_t timeofday0   = (_ticks_t) 0;
#endif


/* The log reader... */

PyDoc_STRVAR(logreader_close__doc__,
"close()\n"
"Close the log file, preventing additional records from being read.");

static PyObject *
logreader_close(LogReaderObject *self, PyObject *args)
{
    if (self->logfp != NULL) {
        fclose(self->logfp);
        self->logfp = NULL;
    }
    Py_INCREF(Py_None);

    return Py_None;
}

PyDoc_STRVAR(logreader_fileno__doc__,
"fileno() -> file descriptor\n"
"Returns the file descriptor for the log file, if open.\n"
"Raises ValueError if the log file is closed.");

static PyObject *
logreader_fileno(LogReaderObject *self)
{
    if (self->logfp == NULL) {
        PyErr_SetString(PyExc_ValueError,
                        "logreader's file object already closed");
        return NULL;
    }
    return PyInt_FromLong(fileno(self->logfp));
}


/* Log File Format
 * ---------------
 *
 * The log file consists of a sequence of variable-length records.
 * Each record is identified with a record type identifier in two
 * bits of the first byte.  The two bits are the "least significant"
 * bits of the byte.
 *
 * Low bits:    Opcode:        Meaning:
 *       0x00         ENTER     enter a frame
 *       0x01          EXIT     exit a frame
 *       0x02        LINENO     execution moved onto a different line
 *       0x03         OTHER     more bits are needed to deecode
 *
 * If the type is OTHER, the record is not packed so tightly, and the
 * remaining bits are used to disambiguate the record type.  These
 * records are not used as frequently so compaction is not an issue.
 * Each of the first three record types has a highly tailored
 * structure that allows it to be packed tightly.
 *
 * The OTHER records have the following identifiers:
 *
 * First byte:  Opcode:        Meaning:
 *       0x13      ADD_INFO     define a key/value pair
 *       0x23   DEFINE_FILE     define an int->filename mapping
 *       0x33    LINE_TIMES     indicates if LINENO events have tdeltas
 *       0x43   DEFINE_FUNC     define a (fileno,lineno)->funcname mapping
 *       0x53   FRAME_TIMES     indicates if ENTER/EXIT events have tdeltas
 *
 * Packed Integers
 *
 * "Packed integers" are non-negative integer values encoded as a
 * sequence of bytes.  Each byte is encoded such that the most
 * significant bit is set if the next byte is also part of the
 * integer.  Each byte provides bits to the least-significant end of
 * the result; the accumulated value must be shifted up to place the
 * new bits into the result.
 *
 * "Modified packed integers" are packed integers where only a portion
 * of the first byte is used.  In the rest of the specification, these
 * are referred to as "MPI(n,name)", where "n" is the number of bits
 * discarded from the least-signicant positions of the byte, and
 * "name" is a name being given to those "discarded" bits, since they
 * are a field themselves.
 *
 * ENTER records:
 *
 *      MPI(2,type)  fileno          -- type is 0x00
 *      PI           lineno
 *      PI           tdelta          -- iff frame times are enabled
 *
 * EXIT records
 *
 *      MPI(2,type)  tdelta          -- type is 0x01; tdelta will be 0
 *                                      if frame times are disabled
 *
 * LINENO records
 *
 *      MPI(2,type)  lineno          -- type is 0x02
 *      PI           tdelta          -- iff LINENO includes it
 *
 * ADD_INFO records
 *
 *      BYTE         type            -- always 0x13
 *      PI           len1            -- length of first string
 *      BYTE         string1[len1]   -- len1 bytes of string data
 *      PI           len2            -- length of second string
 *      BYTE         string2[len2]   -- len2 bytes of string data
 *
 * DEFINE_FILE records
 *
 *      BYTE         type            -- always 0x23
 *      PI           fileno
 *      PI           len             -- length of filename
 *      BYTE         filename[len]   -- len bytes of string data
 *
 * DEFINE_FUNC records
 *
 *      BYTE         type            -- always 0x43
 *      PI           fileno
 *      PI           lineno
 *      PI           len             -- length of funcname
 *      BYTE         funcname[len]   -- len bytes of string data
 *
 * LINE_TIMES records
 *
 * This record can be used only before the start of ENTER/EXIT/LINENO
 * records.  If have_tdelta is true, LINENO records will include the
 * tdelta field, otherwise it will be omitted.  If this record is not
 * given, LINENO records will not contain the tdelta field.
 *
 *      BYTE         type            -- always 0x33
 *      BYTE         have_tdelta     -- 0 if LINENO does *not* have
 *                                      timing information
 * FRAME_TIMES records
 *
 * This record can be used only before the start of ENTER/EXIT/LINENO
 * records.  If have_tdelta is true, ENTER and EXIT records will
 * include the tdelta field, otherwise it will be omitted.  If this
 * record is not given, ENTER and EXIT records will contain the tdelta
 * field.
 *
 *      BYTE         type            -- always 0x53
 *      BYTE         have_tdelta     -- 0 if ENTER/EXIT do *not* have
 *                                      timing information
 */

#define WHAT_ENTER        0x00
#define WHAT_EXIT         0x01
#define WHAT_LINENO       0x02
#define WHAT_OTHER        0x03  /* only used in decoding */
#define WHAT_ADD_INFO     0x13
#define WHAT_DEFINE_FILE  0x23
#define WHAT_LINE_TIMES   0x33
#define WHAT_DEFINE_FUNC  0x43
#define WHAT_FRAME_TIMES  0x53

#define ERR_NONE          0
#define ERR_EOF          -1
#define ERR_EXCEPTION    -2
#define ERR_BAD_RECTYPE  -3

#define PISIZE            (sizeof(int)      + 1)  /* 7-bit bytes */
#define LPISIZE           (sizeof(_ticks_t) + 2)  /* 7-bit bytes */
#define MPISIZE           (PISIZE  + 1)
#define MLPISIZE          (LPISIZE + 2)  /* XXX or 1 */

/* Maximum size of "normal" events -- nothing that contains string data */
#define MAXEVENTSIZE      (MPISIZE + PISIZE + LPISIZE)

#define MODSIZE    2    /* in bits */
#define MODBITS  ( 7  - MODSIZE)
#define MODMASK  ((1 << MODBITS) - 1)

static int
unpack_bytes(LogReaderObject *self, char* bytes, size_t n)
{
    int c;
    int i = 0;
    do {
        c = fgetc(self->logfp);
        if (c == EOF || n <= (size_t) i) {
            return ERR_EOF;
        }
        bytes[i++] = c & 0x7F;
    } while (c & 0x80);
    return i;
}

/* Unpack a packed integer; if "discard" is non-zero, unpack
 * a modified packed integer with "discard" discarded bits.
 */
static int
unpack_packed_int(LogReaderObject *self, int *pvalue, int modified)
{
    char b[MPISIZE+2];
    int i = unpack_bytes(self, b, sizeof(b));
    if (i > 0) {
        int v = (int) b[--i];
        while (i-- > 0) {
            v <<= 7;
            v |= (int) b[i];
        }
       *pvalue = modified ? (v >> MODSIZE) : v;
        return 0;
    }
    return i;
}

static int
unpack_packed_ticks(LogReaderObject *self, _ticks_t *pvalue, int modified)
{
    char b[MLPISIZE+2];
    int i = unpack_bytes(self, b, sizeof(b));
    if (i > 0) {
       _ticks_t v = (_ticks_t) b[--i];
        while (i-- > 0) {
            v <<= 7;
            v |= (_ticks_t) b[i];
        }
       *pvalue = modified ? (v >> MODSIZE) : v;
        return 0;
    }
    return i;
}

/* Unpack a string, which is encoded as a packed integer giving the
 * length of the string, followed by the string data.
 */
static int
unpack_string(LogReaderObject *self, PyObject **pvalue)
{
    int i;
    char str[256];
    char *buf = str;

    int len = 0;
    int err = unpack_packed_int(self, &len, 0);
    if (err) {
        return err;
    }
    if (sizeof(str) < (size_t) len) {
        buf = (char *) malloc((size_t) len);
        if (!buf) {
            PyErr_NoMemory();
            return ERR_EXCEPTION;
        }
    }

    for (i=0; i < len; i++) {
        int ch = fgetc(self->logfp);
        buf[i] = ch;
        if (ch == EOF) {
            err = ERR_EOF;
            goto finally;
        }
    }
    *pvalue = PyString_FromStringAndSize(buf, len);
    if (*pvalue == NULL) {
        err = ERR_EXCEPTION;
    }

 finally:
    if (buf != str) {
        free(buf);
    }
    return err;
}


static int
unpack_add_info(LogReaderObject *self)
{
    PyObject *key;
    PyObject *value = NULL;
    int err;

    err = unpack_string(self, &key);
    if (!err) {
        err = unpack_string(self, &value);
        if (err)
            Py_DECREF(key);
        else {
            PyObject *list = PyDict_GetItem(self->info, key);
            if (list == NULL) {
                list = PyList_New(0);
                if (list == NULL) {
                    err = ERR_EXCEPTION;
                    goto finally;
                }
                if (PyDict_SetItem(self->info, key, list)) {
                    Py_DECREF(list);
                    err = ERR_EXCEPTION;
                    goto finally;
                }
                Py_DECREF(list);
            }
            if (PyList_Append(list, value))
                err = ERR_EXCEPTION;
        }
    }
 finally:
    Py_XDECREF(key);
    Py_XDECREF(value);
    return err;
}


static void
eof_error(LogReaderObject *self)
{
    fclose(self->logfp);
    self->logfp = NULL;
    PyErr_SetString(PyExc_EOFError,
                    "end of file with incomplete profile record");
}

static PyObject *
logreader_tp_iternext(LogReaderObject *self)
{
    int c;
    int what;
    int err = ERR_NONE;
    int lineno = -1;
    int fileno = -1;
   _ticks_t tdelta = (_ticks_t) ~0;  /* -1 */
    PyObject *s1 = NULL, *s2 = NULL;
    PyObject *result = NULL;
#if 0
    unsigned char b0, b1;
#endif

    if (self->logfp == NULL) {
        PyErr_SetString(ProfilerError,
                        "cannot iterate over closed LogReader object");
        return NULL;
    }

restart:
    /* decode the record type */
    if ((c = fgetc(self->logfp)) == EOF) {
        fclose(self->logfp);
        self->logfp = NULL;
        return NULL;
    }
    what = c & WHAT_OTHER;
    if (what == WHAT_OTHER)
        what = c; /* need all the bits for type */
    else
        ungetc(c, self->logfp); /* type byte includes packed int */

    switch (what) {
    case WHAT_ENTER:
        err = unpack_packed_int(self, &fileno, MODSIZE);
        if (!err) {
            err = unpack_packed_int(self, &lineno, 0);
            if (!err && self->frametimings)
                err = unpack_packed_ticks(self, &tdelta, 0);
        }
        break;
    case WHAT_EXIT:
        err = unpack_packed_ticks(self, &tdelta, MODSIZE);
        break;
    case WHAT_LINENO:
        err = unpack_packed_int(self, &lineno, MODSIZE);
        if (!err && self->linetimings)
            err = unpack_packed_ticks(self, &tdelta, 0);
        break;
    case WHAT_ADD_INFO:
        err = unpack_add_info(self);
        break;
    case WHAT_DEFINE_FILE:
        err = unpack_packed_int(self, &fileno, 0);
        if (!err) {
            err = unpack_string(self, &s1);
            if (!err) {
                Py_INCREF(Py_None);
                s2 = Py_None;
            }
        }
        break;
    case WHAT_DEFINE_FUNC:
        err = unpack_packed_int(self, &fileno, 0);
        if (!err) {
            err = unpack_packed_int(self, &lineno, 0);
            if (!err)
                err = unpack_string(self, &s1);
        }
        break;
    case WHAT_LINE_TIMES:
        if ((c = fgetc(self->logfp)) == EOF)
            err = ERR_EOF;
        else {
            self->linetimings = c ? 1 : 0;
            goto restart;
        }
        break;
    case WHAT_FRAME_TIMES:
        if ((c = fgetc(self->logfp)) == EOF)
            err = ERR_EOF;
        else {
            self->frametimings = c ? 1 : 0;
            goto restart;
        }
        break;
    default:
        err = ERR_BAD_RECTYPE;
    }
    if (err == ERR_BAD_RECTYPE) {
        PyErr_SetString(PyExc_ValueError,
                        "unknown record type in log file");
    }
    else if (err == ERR_EOF) {
        eof_error(self);
    }
    else if (!err) {
        result = PyTuple_New(4);
        if (result == NULL)
            return NULL;
        PyTuple_SET_ITEM(result, 0, PyInt_FromLong((long) what));
        PyTuple_SET_ITEM(result, 2, PyInt_FromLong((long) fileno));
#ifdef MS_WINDOWS
        PyTuple_SET_ITEM(result, 1, (s1 ? s1 : PyLong_FromLongLong(tdelta)));
#elif _TICKSIZE == 2
        PyTuple_SET_ITEM(result, 1, (s1 ? s1 : PyLong_FromUnsignedLongLong(tdelta)));
#elif _TICKSIZE == 1
        PyTuple_SET_ITEM(result, 1, (s1 ? s1 : PyLong_FromUnsignedLong(tdelta)));
#else
        PyTuple_SET_ITEM(result, 1, (s1 ? s1 : PyInt_FromLong((long) tdelta)));
#endif
        PyTuple_SET_ITEM(result, 3, (s2 ? s2 : PyInt_FromLong((long) lineno)));
    }
    /* The only other case is err == ERR_EXCEPTION, in which case the
     * exception is already set.
     */
#if 0
    b0 = self->buffer[self->index];
    b1 = self->buffer[self->index + 1];
    if (b0 & 1) {
        /* This is a line-number event. */
        what = PyTrace_LINE;
        lineno = ((b0 & ~1) << 7) + b1;
        self->index += 2;
    }
    else {
        what = (b0 & 0x0E) >> 1;
        tdelta = ((b0 & 0xF0) << 4) + b1;
        if (what == PyTrace_CALL) {
            /* we know there's a 2-byte file ID & 2-byte line number */
            fileno = ((self->buffer[self->index + 2] << 8)
                      + self->buffer[self->index + 3]);
            lineno = ((self->buffer[self->index + 4] << 8)
                      + self->buffer[self->index + 5]);
            self->index += 6;
        }
        else
            self->index += 2;
    }
#endif
    return result;
}

static void
logreader_dealloc(LogReaderObject *self)
{
    if (self->logfp != NULL) {
        fclose(self->logfp);
        self->logfp = NULL;
    }
    Py_XDECREF(self->info);
    PyObject_Del(self);
}

static PyObject *
logreader_sq_item(LogReaderObject *self, Py_ssize_t index)
{
    PyObject *result = logreader_tp_iternext(self);
    if (result == NULL && !PyErr_Occurred()) {
        PyErr_SetString(PyExc_IndexError, "no more events in log");
        return NULL;
    }
    return result;
}

static void
do_stop(ProfilerObject *self);

static int
flush_data(ProfilerObject *self)
{
    /* Need to dump data to the log file... */
    size_t written = fwrite(self->buffer, 1, self->index, self->logfp);
    if (written == (size_t)self->index)
        self->index = 0;
    else {
        memmove(self->buffer, &self->buffer[written],
                self->index - written);
        self->index -= written;
        if (written == 0) {
            char *s = PyString_AsString(self->logfilename);
            PyErr_SetFromErrnoWithFilename(PyExc_IOError, s);
            do_stop(self);
            return -1;
        }
    }
    if (written > 0) {
        if (fflush(self->logfp)) {
            char *s = PyString_AsString(self->logfilename);
            PyErr_SetFromErrnoWithFilename(PyExc_IOError, s);
            do_stop(self);
            return -1;
        }
    }
    return 0;
}

#define FLUSH_DATA(self, n)    (BUFFERSIZE <= ((n) + self->index) && flush_data(self) < 0)

/* Encode an integer.
 */
static inline int
pack_packed_int(ProfilerObject *self, int value)
{
    unsigned char partial;
    do {
        partial = (unsigned char) (value & 0x7F);
        value >>= 7;
        if (value)
            partial |= 0x80;
        self->buffer[self->index] = partial;
        self->index++;
    } while (value);

    return 0;
}

static inline int
pack_packed_ticks(ProfilerObject *self, _ticks_t value)
{
    unsigned char partial;
    do {
        partial = (unsigned char) (value & 0x7F);
        value >>= 7;
        if (value)
            partial |= 0x80;
        self->buffer[self->index] = partial;
        self->index++;
    } while (value);

    return 0;
}

/* Encode a modified packed integer, with a subfield of MODSIZE
 * bits containing value 'what'.  The value of 'what' is not
 * checked to ensure it actually fits in MODSIZE bits.
 */
static inline int
pack_modified_packed_int(ProfilerObject *self, int value, unsigned char what)
{
    int partial = value & MODMASK;
    unsigned char b = (partial != value) ? 0x80 : 0;

    self->buffer[self->index] = b | what | (unsigned char) (partial << MODSIZE);
    self->index++;
    return (b ? pack_packed_int(self, value >> MODBITS) : 0);
}

static inline int
pack_modified_packed_ticks(ProfilerObject *self, _ticks_t value, unsigned char what)
{
   _ticks_t partial = value & (_ticks_t) MODMASK;
    unsigned char b = (partial != value) ? 0x80 : 0;

    self->buffer[self->index] = b | what | (unsigned char) (partial << MODSIZE);
    self->index++;
    return (b ? pack_packed_ticks(self, value >> MODBITS) : 0);
}

static inline int
pack_delta_ticks(ProfilerObject *self, _ticks_t ticks, unsigned char what)
{
    int d = 0;
   _ticks_t p = self->prev_ticks;
    /* time can go backwards on some multiprocessor systems or by NTP */
    if (ticks > p) {
        self->prev_ticks = ticks;
        ticks -= p;
        if (ticks >= (_ticks_t) INT_MAX) {
            return (what ? pack_modified_packed_ticks(self, ticks, what)
                         :          pack_packed_ticks(self, ticks));
        }
        d = (int) ticks;
    }
    return (what ? pack_modified_packed_int(self, d, what)
                 :          pack_packed_int(self, d));
}

static int
pack_string(ProfilerObject *self, const char *s, Py_ssize_t len)
{
    if (FLUSH_DATA(self, len + PISIZE)) {
        return -1;
    }
    assert(len < INT_MAX);
    if (pack_packed_int(self, (int)len) < 0)
        return -1;
    memcpy(self->buffer + self->index, s, len);
    self->index += len;
    return 0;
}

static int
pack_add_info(ProfilerObject *self, const char *s1, const char *s2)
{
    Py_ssize_t len1 = strlen(s1);
    Py_ssize_t len2 = strlen(s2);

    if (FLUSH_DATA(self, len1 + len2 + PISIZE*2)) {
        return -1;
    }
    self->buffer[self->index] = WHAT_ADD_INFO;
    self->index++;
    if (pack_string(self, s1, len1) < 0)
        return -1;
    return pack_string(self, s2, len2);
}

static int
pack_define_file(ProfilerObject *self, int fileno, const char *filename)
{
    Py_ssize_t len = strlen(filename);

    if (FLUSH_DATA(self, len + PISIZE*2 + 1)) {
        return -1;
    }
    self->buffer[self->index] = WHAT_DEFINE_FILE;
    self->index++;
    if (pack_packed_int(self, fileno) < 0)
        return -1;
    return pack_string(self, filename, len);
}

static int
pack_define_func(ProfilerObject *self, int fileno, int lineno,
                 const char *funcname)
{
    Py_ssize_t len = strlen(funcname);

    if (FLUSH_DATA(self, len + PISIZE*3 + 1)) {
        return -1;
    }
    self->buffer[self->index] = WHAT_DEFINE_FUNC;
    self->index++;
    if (pack_packed_int(self, fileno) < 0)
        return -1;
    if (pack_packed_int(self, lineno) < 0)
        return -1;
    return pack_string(self, funcname, len);
}

static int
pack_line_times(ProfilerObject *self)
{
    if (FLUSH_DATA(self, 2)) {
        return -1;
    }
    self->buffer[self->index] = WHAT_LINE_TIMES;
    self->buffer[self->index + 1] = self->linetimings ? 1 : 0;
    self->index += 2;
    return 0;
}

static int
pack_frame_times(ProfilerObject *self)
{
    if (FLUSH_DATA(self, 2)) {
        return -1;
    }
    self->buffer[self->index] = WHAT_FRAME_TIMES;
    self->buffer[self->index + 1] = self->frametimings ? 1 : 0;
    self->index += 2;
    return 0;
}

static inline int
pack_enter(ProfilerObject *self, int fileno, int lineno, _ticks_t ticks)
{
    if (FLUSH_DATA(self, MPISIZE + PISIZE + LPISIZE)) {
        return -1;
    }
    pack_modified_packed_int(self, fileno, WHAT_ENTER);
    pack_packed_int(self, lineno);
    if (self->frametimings)
        return pack_delta_ticks(self, ticks, 0);
    return 0;
}

static inline int
pack_exit(ProfilerObject *self, _ticks_t ticks)
{
    if (FLUSH_DATA(self, MLPISIZE)) {
        return -1;
    }
    if (self->frametimings)
        return pack_delta_ticks(self, ticks, WHAT_EXIT);
    self->buffer[self->index] = WHAT_EXIT;
    self->index++;
    return 0;
}

static inline int
pack_lineno(ProfilerObject *self, int lineno, _ticks_t ticks)
{
    if (FLUSH_DATA(self, MPISIZE + LPISIZE)) {
        return -1;  /* 0? */
    }
    if (pack_modified_packed_int(self, lineno, WHAT_LINENO) < 0)
        return -1;
    if (self->linetimings)
       return pack_delta_ticks(self, ticks, 0);
    return 0;
}

static inline int
get_fileno(ProfilerObject *self, PyCodeObject *fcode)
{
    /* This is only used for ENTER events. */

    PyObject *obj;
    PyObject *dict;
    int fileno;

    obj = PyDict_GetItem(self->filemap, fcode->co_filename);
    if (obj == NULL) {
        /* first sighting of this file */
        dict = PyDict_New();
        if (dict == NULL) {
            return -1;
        }
        fileno = self->next_fileno;
        obj = Py_BuildValue("iN", fileno, dict);
        if (obj == NULL) {
            return -1;
        }
        if (PyDict_SetItem(self->filemap, fcode->co_filename, obj)) {
            Py_DECREF(obj);
            return -1;
        }
        self->next_fileno++;
        Py_DECREF(obj);
        if (pack_define_file(self, fileno,
                             PyString_AS_STRING(fcode->co_filename)) < 0)
            return -1;
    }
    else {
        /* already know this ID */
        fileno = PyInt_AS_LONG(PyTuple_GET_ITEM(obj, 0));
        dict = PyTuple_GET_ITEM(obj, 1);
    }
    /* make sure we save a function name for this (fileno, lineno) */
    obj = PyInt_FromLong(fcode->co_firstlineno);
    if (obj == NULL) {
        /* We just won't have it saved; too bad. */
        PyErr_Clear();
    }
    else {
        PyObject *name = PyDict_GetItem(dict, obj);
        if (name == NULL) {
            if (pack_define_func(self, fileno, fcode->co_firstlineno,
                                 PyString_AS_STRING(fcode->co_name)) < 0) {
                Py_DECREF(obj);
                return -1;
            }
            if (PyDict_SetItem(dict, obj, fcode->co_name)) {
                Py_DECREF(obj);
                return -1;
            }
        }
        Py_DECREF(obj);
    }
    return fileno;
}


/* The workhorse:  the profiler callback function. */

static int
tracer_callback(ProfilerObject *self, PyFrameObject *frame, int what,
                PyObject *arg)
{
    int fileno;

    switch (what) {
    case PyTrace_CALL:
        fileno = get_fileno(self, frame->f_code);
        if (fileno < 0)
            return -1;
        return pack_enter(self, fileno, frame->f_code->co_firstlineno,
                          self->frametimings ? _gethrticks() : (_ticks_t) 0);

    case PyTrace_RETURN:
        return pack_exit(self, self->frametimings ? _gethrticks() : (_ticks_t) 0);

    case PyTrace_LINE:  /* we only get these events if we asked for them */
        return pack_lineno(self, frame->f_lineno,
                           self->linetimings ? _gethrticks() : (_ticks_t) 0);

    default:
        /* ignore PyTrace_EXCEPTION */
        break;
    }
    return 0;
}


/* A couple of useful helper functions. */

static void
calibrate (int cycles)
{
    int i;
    long d;
   _ticks_t t1, t2;
#ifdef MS_WINDOWS
    LARGE_INTEGER f;
    QueryPerformanceFrequency(&f);
    timer_frequency = (double) f.QuadPart;
    if (!timer_resolution) {
        t1 = _gethrticks();
        do {
            t2 = _gethrticks();
        } while (t2 <= t1);
        timer_resolution = (long) (t2 - t1);  /* ticks */
    }
#else
    for (i = cycles; i-- > 0; ) {
        t1 = _gettimeofday();
        do {
            t2 = _gettimeofday();
        } while (t2 <= t1);
        d = (long) (t2 - t1);
        if (timeofday_diff > d || !timeofday_diff) {
            timeofday_diff = d;
        }
    }
# if _RUSAGE
    for (i = cycles; i-- > 0; ) {
        t1 = _getrusage();
        do {
            t2 = _getrusage();
        } while (t2 <= t1);
        d = (long) (t2 - t1);
        if (rusage_diff > d || !rusage_diff) {
            rusage_diff = d;
        }
    }
# else
    rusage_diff = -1L;  /* n/a */
# endif
# if _TICKSIZE
    if (!timer_resolution) {
        t1 = _gethrticks();
        do {
            t2 = _gethrticks();
        } while (t2 <= t1);
        timer_resolution = (long) (t2 - t1);  /* ticks */
    }
    /* determine hires timer frequency by waiting at least
       0.25 secs since the very start and divide the number
       of ticks since by the actual elapsed time in seconds */
    t1 = timeofday0 + (_ticks_t) 250000;  /* 0.25 secs */
    do {
        t2 = _gettimeofday();
    } while (t2 <= t1);
    t1 = _gethrticks();
    timer_frequency = 1.0e6 * (double) (t1 - ticks0) / (double) (t2 - timeofday0);
# else
    timer_resolution = timeofday_diff;
    /* clock ticks may not be in usecs
       but _gettimeofday values are */
    timer_frequency  = 1000000.0;  /* usecs */
# endif
#endif
}

static void
do_start(ProfilerObject *self)
{
    self->active = 1;
    self->prev_ticks = _gethrticks();
    if (self->lineevents)
        PyEval_SetTrace((Py_tracefunc) tracer_callback, (PyObject *)self);
    else
        PyEval_SetProfile((Py_tracefunc) tracer_callback, (PyObject *)self);
}

static void
do_stop(ProfilerObject *self)
{
    if (self->active) {
        self->active = 0;
        if (self->lineevents)
            PyEval_SetTrace(NULL, NULL);
        else
            PyEval_SetProfile(NULL, NULL);
    }
    if (self->index > 0) {
        /* Best effort to dump out any remaining data. */
        flush_data(self);
    }
}

static int
is_available(ProfilerObject *self)
{
    if (self->active) {
        PyErr_SetString(ProfilerError, "profiler already active");
        return 0;
    }
    if (self->logfp == NULL) {
        PyErr_SetString(ProfilerError, "profiler already closed");
        return 0;
    }
    return 1;
}


/* Profiler object interface methods. */

PyDoc_STRVAR(addinfo__doc__,
"addinfo(key, value)\n"
"Insert an ADD_INFO record into the log.");

static PyObject *
profiler_addinfo(ProfilerObject *self, PyObject *args)
{
    PyObject *result = NULL;
    char *key, *value;

    if (PyArg_ParseTuple(args, "ss:addinfo", &key, &value)) {
        if (self->logfp == NULL)
            PyErr_SetString(ProfilerError, "profiler already closed");
        else {
            if (pack_add_info(self, key, value) == 0) {
                result = Py_None;
                Py_INCREF(result);
            }
        }
    }
    return result;
}

PyDoc_STRVAR(close__doc__,
"close()\n"
"Shut down this profiler and close the log files, even if its active.");

static PyObject *
profiler_close(ProfilerObject *self)
{
    do_stop(self);
    if (self->logfp != NULL) {
        fclose(self->logfp);
        self->logfp = NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

#define fileno__doc__ logreader_fileno__doc__

static PyObject *
profiler_fileno(ProfilerObject *self)
{
    if (self->logfp == NULL) {
        PyErr_SetString(PyExc_ValueError,
                        "profiler's file object already closed");
        return NULL;
    }
    return PyInt_FromLong(fileno(self->logfp));
}

PyDoc_STRVAR(runcall__doc__,
"runcall(callable[, args[, kw]]) -> callable()\n"
"Profile a specific function call, returning the result of that call.");

static PyObject *
profiler_runcall(ProfilerObject *self, PyObject *args)
{
    PyObject *result = NULL;
    PyObject *callargs = NULL;
    PyObject *callkw = NULL;
    PyObject *callable;

    if (PyArg_UnpackTuple(args, "runcall", 1, 3,
                         &callable, &callargs, &callkw)) {
        if (is_available(self)) {
            do_start(self);
            result = PyEval_CallObjectWithKeywords(callable, callargs, callkw);
            do_stop(self);
        }
    }
    return result;
}

PyDoc_STRVAR(runcode__doc__,
"runcode(code, globals[, locals])\n"
"Execute a code object while collecting profile data.  If locals is\n"
"omitted, globals is used for the locals as well.");

static PyObject *
profiler_runcode(ProfilerObject *self, PyObject *args)
{
    PyObject *result = NULL;
    PyCodeObject *code;
    PyObject *globals;
    PyObject *locals = NULL;

    if (PyArg_ParseTuple(args, "O!O!|O:runcode",
                         &PyCode_Type, &code,
                         &PyDict_Type, &globals,
                         &locals)) {
        if (is_available(self)) {
            if (locals == NULL || locals == Py_None)
                locals = globals;
            else if (!PyDict_Check(locals)) {
                PyErr_SetString(PyExc_TypeError,
                                "locals must be a dictionary or None");
                return NULL;
            }
            do_start(self);
            result = PyEval_EvalCode(code, globals, locals);
            do_stop(self);
#if 0
            if (!PyErr_Occurred()) {
                result = Py_None;
                Py_INCREF(result);
            }
#endif
        }
    }
    return result;
}

PyDoc_STRVAR(start__doc__,
"start()\n"
"Install this profiler for the current thread.");

static PyObject *
profiler_start(ProfilerObject *self, PyObject *args)
{
    PyObject *result = NULL;

    if (is_available(self)) {
        do_start(self);
        result = Py_None;
        Py_INCREF(result);
    }
    return result;
}

PyDoc_STRVAR(stop__doc__,
"stop()\n"
"Remove this profiler from the current thread.");

static PyObject *
profiler_stop(ProfilerObject *self, PyObject *args)
{
    PyObject *result = NULL;

    if (!self->active)
        PyErr_SetString(ProfilerError, "profiler not active");
    else {
        do_stop(self);
        result = Py_None;
        Py_INCREF(result);
    }
    return result;
}


/* Python API support. */

static void
profiler_dealloc(ProfilerObject *self)
{
    do_stop(self);
    if (self->logfp != NULL)
        fclose(self->logfp);
    Py_XDECREF(self->filemap);
    Py_XDECREF(self->logfilename);
    PyObject_Del((PyObject *)self);
}

static PyMethodDef profiler_methods[] = {
    {"addinfo", (PyCFunction)profiler_addinfo, METH_VARARGS, addinfo__doc__},
    {"close",   (PyCFunction)profiler_close,   METH_NOARGS,  close__doc__},
    {"fileno",  (PyCFunction)profiler_fileno,  METH_NOARGS,  fileno__doc__},
    {"runcall", (PyCFunction)profiler_runcall, METH_VARARGS, runcall__doc__},
    {"runcode", (PyCFunction)profiler_runcode, METH_VARARGS, runcode__doc__},
    {"start",   (PyCFunction)profiler_start,   METH_NOARGS,  start__doc__},
    {"stop",    (PyCFunction)profiler_stop,    METH_NOARGS,  stop__doc__},
    {NULL, NULL}
};

static PyMemberDef profiler_members[] = {
    {"frametimings", T_LONG, offsetof(ProfilerObject, linetimings), READONLY},
    {"lineevents",   T_LONG, offsetof(ProfilerObject, lineevents), READONLY},
    {"linetimings",  T_LONG, offsetof(ProfilerObject, linetimings), READONLY},
    {NULL}
};

static PyObject *
profiler_get_closed(ProfilerObject *self, void *closure)
{
    PyObject *result = (self->logfp == NULL) ? Py_True : Py_False;
    Py_INCREF(result);
    return result;
}

static PyGetSetDef profiler_getsets[] = {
    {"closed", (getter)profiler_get_closed, NULL,
     PyDoc_STR("True if the profiler's output file has already been closed.")},
    {NULL}
};


PyDoc_STRVAR(profiler_object__doc__,
"High-performance profiler object.\n"
"\n"
"Methods:\n"
"\n"
"close():      Stop the profiler and close the log files.\n"
"fileno():     Returns the file descriptor of the log file.\n"
"runcall():    Run a single function call with profiling enabled.\n"
"runcode():    Execute a code object with profiling enabled.\n"
"start():      Install the profiler and return.\n"
"stop():       Remove the profiler.\n"
"\n"
"Attributes (read-only):\n"
"\n"
"closed:       True if the profiler has already been closed.\n"
"frametimings: True if ENTER/EXIT events collect timing information.\n"
"lineevents:   True if line events are reported to the profiler.\n"
"linetimings:  True if line events collect timing information.");

static PyTypeObject ProfilerType = {
    PyObject_HEAD_INIT(NULL)
    0,					/* ob_size		*/
    "_hotshot.ProfilerType",		/* tp_name		*/
    (int) sizeof(ProfilerObject),	/* tp_basicsize		*/
    0,					/* tp_itemsize		*/
    (destructor)profiler_dealloc,	/* tp_dealloc		*/
    0,					/* tp_print		*/
    0,					/* tp_getattr		*/
    0,					/* tp_setattr		*/
    0,					/* tp_compare		*/
    0,					/* tp_repr		*/
    0,					/* tp_as_number		*/
    0,					/* tp_as_sequence	*/
    0,					/* tp_as_mapping	*/
    0,					/* tp_hash		*/
    0,					/* tp_call		*/
    0,					/* tp_str		*/
    PyObject_GenericGetAttr,		/* tp_getattro		*/
    0,					/* tp_setattro		*/
    0,					/* tp_as_buffer		*/
    Py_TPFLAGS_DEFAULT,			/* tp_flags		*/
    profiler_object__doc__,		/* tp_doc		*/
    0,					/* tp_traverse		*/
    0,					/* tp_clear		*/
    0,					/* tp_richcompare	*/
    0,					/* tp_weaklistoffset	*/
    0,					/* tp_iter		*/
    0,					/* tp_iternext		*/
    profiler_methods,			/* tp_methods		*/
    profiler_members,			/* tp_members		*/
    profiler_getsets,			/* tp_getset		*/
    0,					/* tp_base		*/
    0,					/* tp_dict		*/
    0,					/* tp_descr_get		*/
    0,					/* tp_descr_set		*/
};


static PyMethodDef logreader_methods[] = {
    {"close",   (PyCFunction)logreader_close,  METH_NOARGS,
     logreader_close__doc__},
    {"fileno",  (PyCFunction)logreader_fileno, METH_NOARGS,
     logreader_fileno__doc__},
    {NULL, NULL}
};

static PyMemberDef logreader_members[] = {
    {"info", T_OBJECT, offsetof(LogReaderObject, info), RO,
     PyDoc_STR("Dictionary mapping informational keys to lists of values.")},
    {NULL}
};


PyDoc_STRVAR(logreader__doc__,
"logreader(filename) --> log-iterator\n\
Create a log-reader for the timing information file.");

static PySequenceMethods logreader_as_sequence = {
    0,					/* sq_length */
    0,					/* sq_concat */
    0,					/* sq_repeat */
    (ssizeargfunc)logreader_sq_item,	/* sq_item */
    0,					/* sq_slice */
    0,					/* sq_ass_item */
    0,					/* sq_ass_slice */
    0,					/* sq_contains */
    0,					/* sq_inplace_concat */
    0,					/* sq_inplace_repeat */
};

static PyObject *
logreader_get_closed(LogReaderObject *self, void *closure)
{
    PyObject *result = (self->logfp == NULL) ? Py_True : Py_False;
    Py_INCREF(result);
    return result;
}

static PyGetSetDef logreader_getsets[] = {
    {"closed", (getter)logreader_get_closed, NULL,
     PyDoc_STR("True if the logreader's input file has already been closed.")},
    {NULL}
};

static PyTypeObject LogReaderType = {
    PyObject_HEAD_INIT(NULL)
    0,					/* ob_size		*/
    "_hotshot.LogReaderType",		/* tp_name		*/
    (int) sizeof(LogReaderObject),	/* tp_basicsize		*/
    0,					/* tp_itemsize		*/
    (destructor)logreader_dealloc,	/* tp_dealloc		*/
    0,					/* tp_print		*/
    0,					/* tp_getattr		*/
    0,					/* tp_setattr		*/
    0,					/* tp_compare		*/
    0,					/* tp_repr		*/
    0,					/* tp_as_number		*/
    &logreader_as_sequence,		/* tp_as_sequence	*/
    0,					/* tp_as_mapping	*/
    0,					/* tp_hash		*/
    0,					/* tp_call		*/
    0,					/* tp_str		*/
    PyObject_GenericGetAttr,		/* tp_getattro		*/
    0,					/* tp_setattro		*/
    0,					/* tp_as_buffer		*/
    Py_TPFLAGS_DEFAULT,			/* tp_flags		*/
    logreader__doc__,			/* tp_doc		*/
    0,					/* tp_traverse		*/
    0,					/* tp_clear		*/
    0,					/* tp_richcompare	*/
    0,					/* tp_weaklistoffset	*/
    PyObject_SelfIter,			/* tp_iter		*/
    (iternextfunc)logreader_tp_iternext,/* tp_iternext		*/
    logreader_methods,			/* tp_methods		*/
    logreader_members,			/* tp_members		*/
    logreader_getsets,			/* tp_getset		*/
    0,					/* tp_base		*/
    0,					/* tp_dict		*/
    0,					/* tp_descr_get		*/
    0,					/* tp_descr_set		*/
};

static PyObject *
hotshot_logreader(PyObject *unused, PyObject *args)
{
    LogReaderObject *self = NULL;
    char *filename;
    int c;
    int err = 0;

    if (PyArg_ParseTuple(args, "s:logreader", &filename)) {
        self = PyObject_New(LogReaderObject, &LogReaderType);
        if (self != NULL) {
            self->frametimings = 1;
            self->linetimings = 0;
            self->info = NULL;
            self->logfp = fopen(filename, "rb");
            if (self->logfp == NULL) {
                PyErr_SetFromErrnoWithFilename(PyExc_IOError, filename);
                Py_DECREF(self);
                self = NULL;
                goto finally;
            }
            self->info = PyDict_New();
            if (self->info == NULL) {
                Py_DECREF(self);
                goto finally;
            }
            /* read initial info */
            for (;;) {
                if ((c = fgetc(self->logfp)) == EOF) {
                    eof_error(self);
                    break;
                }
                if (c != WHAT_ADD_INFO) {
                    ungetc(c, self->logfp);
                    break;
                }
                err = unpack_add_info(self);
                if (err) {
                    if (err == ERR_EOF)
                        eof_error(self);
                    else
                        PyErr_SetString(PyExc_RuntimeError,
                                        "unexpected error");
                    break;
                }
            }
        }
    }
 finally:
    return (PyObject *) self;
}


/* Return a string that represents the version number
 * without the extra cruft added by revision control,
 * even if the right options were given to "cvs export"
 * to make it not include the extra cruft.
 */
static char *
get_version_string(char *str, size_t n)
{
    static char *rcsid = "$Revision: 51218 $";
    char *rev = rcsid;
    int i = 0;

    while (*rev && !isdigit(Py_CHARMASK(*rev))) {
        ++rev;
    }
    while (*rev && *rev != ' ' && n-- > 1) {
        str[i++] = *rev++;
    }
    str[i] = '\0';
    return str;
}

/* Write out a RFC 822-style header with various useful bits of
 * information to make the output easier to manage.
 */
static int
write_header(ProfilerObject *self)
{
    char *buffer;
    char str[PATH_MAX];
    PyObject *temp;
    Py_ssize_t i, len;

    pack_add_info(self, "hotshot-version",
                  get_version_string(str, sizeof(str)));
    pack_add_info(self, "requested-frame-timings",
                  (self->frametimings ? "yes" : "no"));
    pack_add_info(self, "requested-line-events",
                  (self->lineevents ? "yes" : "no"));
    pack_add_info(self, "requested-line-timings",
                  (self->linetimings ? "yes" : "no"));
    pack_add_info(self, "platform", Py_GetPlatform());
    pack_add_info(self, "executable", Py_GetProgramFullPath());
    buffer = (char *) Py_GetVersion();
    if (buffer == NULL)
        PyErr_Clear();
    else
        pack_add_info(self, "executable-version", buffer);

    if (!timeofday_diff) {
        calibrate(1);
    }
    PyOS_snprintf(str, sizeof(str), "%ld", timer_resolution);
    pack_add_info(self, "timer-resolution", str);
    PyOS_snprintf(str, sizeof(str), "%.3f", timer_frequency);
    pack_add_info(self, "timer-frequency", str);
#ifdef MS_WINDOWS
    PyOS_snprintf(str, sizeof(str), "%.0f", timer_frequency);
    pack_add_info(self, "reported-performance-frequency", str);
#else
    PyOS_snprintf(str, sizeof(str), "%ld", timeofday_diff);
    pack_add_info(self, "observed-interval-gettimeofday", str);
    PyOS_snprintf(str, sizeof(str), "%ld", rusage_diff);
    pack_add_info(self, "observed-interval-getrusage", str);
#endif
    pack_add_info(self, "current-directory", getcwd(str, sizeof(str)));

    temp = PySys_GetObject("path");
    if (temp == NULL || !PyList_Check(temp)) {
        PyErr_SetString(PyExc_RuntimeError, "sys.path must be a list");
        return -1;
    }
    len = PyList_GET_SIZE(temp);
    for (i = 0; i < len; ++i) {
        PyObject *item = PyList_GET_ITEM(temp, i);
        buffer = PyString_AsString(item);
        if (buffer == NULL) {
            pack_add_info(self, "sys-path-entry", "<non-string-path-entry>");
            PyErr_Clear();
        }
        else {
            pack_add_info(self, "sys-path-entry", buffer);
        }
    }
    pack_frame_times(self);
    pack_line_times(self);

    return 0;
}

PyDoc_STRVAR(profiler__doc__,
"profiler(logfilename[, lineevents[, linetimes]]) -> profiler\n\
Create a new profiler object.");

static PyObject *
hotshot_profiler(PyObject *unused, PyObject *args)
{
    char *logfilename;
    ProfilerObject *self = NULL;
    int lineevents = 0;
    int linetimings = 1;

    if (PyArg_ParseTuple(args, "s|ii:profiler", &logfilename,
                         &lineevents, &linetimings)) {
        self = PyObject_New(ProfilerObject, &ProfilerType);
        if (self == NULL)
            return NULL;
        self->frametimings = 1;
        self->lineevents = lineevents ? 1 : 0;
        self->linetimings = (lineevents && linetimings) ? 1 : 0;
        self->index = 0;
        self->active = 0;
        self->next_fileno = 0;
        self->logfp = NULL;
        self->logfilename = PyTuple_GET_ITEM(args, 0);
        Py_INCREF(self->logfilename);
        self->filemap = PyDict_New();
        if (self->filemap == NULL) {
            Py_DECREF(self);
            return NULL;
        }
        self->logfp = fopen(logfilename, "wb");
        if (self->logfp == NULL) {
            Py_DECREF(self);
            PyErr_SetFromErrnoWithFilename(PyExc_IOError, logfilename);
            return NULL;
        }
        if (!timeofday_diff) {
            /* Run this several times since sometimes the
             * first doesn't give the lowest values, and
             * we're really trying to determine the lowest.
             */
            calibrate(3);
        }
        if (write_header(self)) {
            /* some error occurred, exception has been set */
            Py_DECREF(self);
            self = NULL;
        }
    }
    return (PyObject *) self;
}

PyDoc_STRVAR(coverage__doc__,
"coverage(logfilename) -> profiler\n\
Returns a profiler that doesn't collect any timing information, which is\n\
useful in building a coverage analysis tool.");

static PyObject *
hotshot_coverage(PyObject *unused, PyObject *args)
{
    char *logfilename;
    PyObject *result = NULL;

    if (PyArg_ParseTuple(args, "s:coverage", &logfilename)) {
        result = hotshot_profiler(unused, args);
        if (result != NULL) {
            ProfilerObject *self = (ProfilerObject *) result;
            self->frametimings = 0;
            self->linetimings = 0;
            self->lineevents = 1;
        }
    }
    return result;
}

PyDoc_VAR(resolution__doc__) = 
#ifdef MS_WINDOWS
PyDoc_STR(
"resolution() -> (performance-counter-ticks, update-frequency)\n"
"Return the resolution of the timer provided by the QueryPerformanceCounter()\n"
"function.  The first value is the smallest observed change, and the second\n"
"is the result of QueryPerformanceFrequency()."
)
#elif _TICKSIZE
PyDoc_STR(
"resolution() -> (timer-resolution, timer-frequency)\n"
"Return the resolution (in ticks) and frequency (in ticks per second)\n"
"of the high-resolution timer."
)
#else
PyDoc_STR(
"resolution() -> (gettimeofday-usecs, getrusage-usecs)\n"
"Return the resolution of the timers provided by the gettimeofday() and\n"
"getrusage() system calls, or -1 if the call is not supported."
)
#endif
;

static PyObject *
hotshot_resolution(PyObject *self, PyObject *unused)
{
#ifdef MS_WINDOWS
    calibrate(1);
    return Py_BuildValue("ld", timer_resolution, timer_frequency);
#elif _TICKSIZE
    calibrate(1);
    return Py_BuildValue("ld", timer_resolution, timer_frequency);
#else
    if (!timeofday_diff) {
        calibrate(3);
    }
    return Py_BuildValue("ll", timeofday_diff, rusage_diff);
#endif
}


static PyMethodDef functions[] = {
    {"coverage",   hotshot_coverage,   METH_VARARGS, coverage__doc__},
    {"profiler",   hotshot_profiler,   METH_VARARGS, profiler__doc__},
    {"logreader",  hotshot_logreader,  METH_VARARGS, logreader__doc__},
    {"resolution", hotshot_resolution, METH_NOARGS,  resolution__doc__},
    {NULL, NULL}
};


void
init_hotshot(void)
{
    char str[64];
    PyObject *module;
    LogReaderType.ob_type = &PyType_Type;
    ProfilerType.ob_type = &PyType_Type;
    module = Py_InitModule("_hotshot", functions);
    if (module != NULL) {
#if _TICKSIZE
        /* get initial clock and hires time values */
        timeofday0 = _gettimeofday();
        ticks0     = _gethrticks();
#endif
        PyModule_AddStringConstant(module, "__version__",
                                   get_version_string(str, sizeof(str)));
        Py_INCREF(&LogReaderType);
        PyModule_AddObject(module, "LogReaderType",
                           (PyObject *)&LogReaderType);
        Py_INCREF(&ProfilerType);
        PyModule_AddObject(module, "ProfilerType",
                           (PyObject *)&ProfilerType);

        if (ProfilerError == NULL)
            ProfilerError = PyErr_NewException("hotshot.ProfilerError",
                                               NULL, NULL);
        if (ProfilerError != NULL) {
            Py_INCREF(ProfilerError);
            PyModule_AddObject(module, "ProfilerError", ProfilerError);
        }
        PyModule_AddIntConstant(module, "WHAT_ENTER", WHAT_ENTER);
        PyModule_AddIntConstant(module, "WHAT_EXIT", WHAT_EXIT);
        PyModule_AddIntConstant(module, "WHAT_LINENO", WHAT_LINENO);
        PyModule_AddIntConstant(module, "WHAT_OTHER", WHAT_OTHER);
        PyModule_AddIntConstant(module, "WHAT_ADD_INFO", WHAT_ADD_INFO);
        PyModule_AddIntConstant(module, "WHAT_DEFINE_FILE", WHAT_DEFINE_FILE);
        PyModule_AddIntConstant(module, "WHAT_DEFINE_FUNC", WHAT_DEFINE_FUNC);
        PyModule_AddIntConstant(module, "WHAT_LINE_TIMES", WHAT_LINE_TIMES);
    }
}
